from pureml_evaluate.policy.performance import Performance
from pureml_evaluate.policy.fairness import FairnessPolicy as Fairness
import numpy as np
from collections import defaultdict


class Grader:
    list_of_performance_metrics = [
    'accuracy',
    'precision',
    'recall' ,
    'f1',
    'confusionmatrix',
    'balancedaccuracyscore',
    'topkaccuracyscore' ,
    'logloss' ,
    'averageprecisionscore',
    'roc_auc' ,
    'brierscoreloss' ,
    'kolmogorovsmirnov' ,
    'wassersteindistance',
    'hellingerdistance' ,
    'linfinitydistance' ,
    'chisquaredstatistic',
    'cramersv',
    'populationstabilityindex' 
    ]
    metrics : list = []
    policy : dict   
    references : any
    predictions : any
    sensitive_features : any
    scores : dict = {}
    categories: dict = {}

    def __init__(self,references,predictions,sensitive_features,policy,metrics):
        self.references = references
        self.predictions = predictions
        self.sensitive_features  = sensitive_features
        self.policy = policy             #{'acccuracy' : 0.8}
        self.metrics = metrics       # ['accuracy']


    def compute(self):
        all_metrics = []
        
        operational_category = {} # To Store the flagged values of metrics like pass/fail
        fairness_category  = {}
        
        operational_scores = {} # To Store the values of metrics 
        fairness_scores = {}

        operational_threshold = {} # To Store Threshold values
        fairness_threshold = {}

        for metric in self.metrics:
            if metric in self.list_of_performance_metrics:
                performance = Performance()
                threshold_value = self.policy[metric]
                threshold = {metric : threshold_value}
                result = performance.compute(list_metrics = [metric],references = self.references,predictions = self.predictions,
                                             list_of_thresholds = threshold,prediction_scores = None)
                
                temp_scores  = {
                    f"{result['risk']}" : f"{result['value']}"
                }
                operational_scores.update(temp_scores)
                temp_category = {
                    f"{result['risk']}" : f"{result['severity']}"
                }
                operational_category.update(temp_category)
                temp_threshold  = {
                    f"{result['risk']}" : f"{threshold_value}"
                }
                operational_threshold.update(temp_threshold)

                

            else:
                fairness = Fairness()
                threshold_value = self.policy[metric]
                threshold = {metric : threshold_value}
                result = fairness.compute(list_metrics = [metric],references = self.references,predictions = self.predictions,
                                          list_of_thresholds = threshold,sensitive_features = self.sensitive_features,prediction_scores = None)
                
                temp_scores  = {
                    f"{result['risk']}" : f"{result['value']}"
                }
                fairness_scores.update(temp_scores)
                temp_category = {
                    f"{result['risk']}" : f"{result['severity']}"
                }
                fairness_category.update(temp_category)    
                temp_threshold  = {
                    f"{result['risk']}" : f"{threshold_value}"
                }
                fairness_threshold.update(temp_threshold)
                

        self.scores = {
                "operational" : operational_category,
                "fairness" : fairness_category,
                "operational_scores" : operational_scores,
                "fairness_scores" : fairness_scores,
                "operational_thresholds" : operational_threshold,
                "fairness_thresholds" : fairness_threshold
            }
        
        return self.scores




from pydantic import BaseModel
from pureml.predictor.predictor import BasePredictor
from pureml.components import dataset
from typing import Any
from importlib import import_module
from rich import print
import requests
from pureml_evaluate.evaluators.evaluator import eval as eval_fn
from pureml.cli.auth import get_auth_headers
from pureml.components import get_org_id
from .grade import Grader
from collections import defaultdict
import numpy as np
from typing import Union
import matplotlib.pyplot as plt



class EvalHelper(BaseModel):  # To pass the requirements to eval in pureml_evaluate
    label_model: str
    label_dataset: str
    policy: dict
    predictor: BasePredictor = None
    predictor_path: str = "predict.py"
    dataset: Any = None
    sensitive_features: Union[None, Any]
    y_true: Any
    y_pred: Any
    y_pred_scores: Any = None
    


    class Config:
        validate_assignment = True
        arbitrary_types_allowed = True

    def load_dataset(self):
        self.dataset = dataset.fetch(self.label_dataset)
        print("[bold green] Succesfully fetched the dataset")

    def load_predictor(self):
        module_path = self.predictor_path.replace(".py", "")
        module_import = import_module(module_path)

        predictor_class = getattr(module_import, "Predictor")

        self.predictor = predictor_class()
        print("[bold green] Succesfully fetched the predictor")

    def load_model(self):
        self.predictor.load_models()
        print("[bold green] Succesfully fetched the model")

    def load_policy(self):
        return list(self.policy.keys())
    
    def load_sensitive_features(self):
        if 'sensitive_features' in self.dataset.keys():
            self.sensitive_features = self.dataset['sensitive_features']
            return self.dataset['sensitive_features']
        else:
            return None

    def load_y_true(self):
        self.y_true = self.dataset["y_test"]
        return self.dataset["y_test"]
    
    def load_y_pred(self):
        self.y_pred = self.predictor.predict(self.dataset["x_test"])
        return self.predictor.predict(self.dataset["x_test"])
    
    def load(self):
        self.load_dataset()
        self.load_predictor()
        self.load_model()
        self.load_policy()
        self.load_sensitive_features()
        self.load_y_true()
        self.load_y_pred()

    def get_y_pred(self):
        return self.predictor.predict(self.dataset["x_test"])

    def get_y_true(self):
        return self.dataset["y_test"]

    def get_sensitive_features(self):
        print(f"Dataset Keys: {self.dataset.keys()}")
        if 'sensitive_features' in self.dataset.keys():
            print(
                f"Data in sensitive_features: {self.dataset['sensitive_features']}")
            return self.dataset['sensitive_features']
        else:
            return None

    def evaluate(self):
        y_pred = self.get_y_pred()
        y_true = self.get_y_true()
        sensitive_features = self.get_sensitive_features()
        metrics = self.load_policy()
        grader = Grader(references=y_true, predictions=y_pred,
                        sensitive_features=sensitive_features, policy=self.policy, metrics=metrics)

        result  = grader.compute()
        # labelmodel = self.label_model.split(':')
        # version = labelmodel[2]
        # model_branch = f"{labelmodel[0]}:{labelmodel[1]}"
        # #print(f"version: {version}")
        # #print(f"model_branch: {model_branch}")
        # formatted_result = {
        #     "model" : f"{model_branch}",
        #     "version" : f"{version}",
        #     "result" : [result]
        # }
        # return formatted_result
        return {'complete' : result}

    def evaluate_subsets(self):
        if self.sensitive_features is None:  # If No Sensitive Features are given
            return 
    
        if self.sensitive_features is not None: 
            subsets = self.give_subsets()

            values_subsets_all = {}

            for subset in subsets:
                values_all = defaultdict(dict)

                key = subset['key']
                y_true = subset['y_true']
                y_pred = subset['y_pred']
                sensitive_features = subset['sensitive_features']
                y_pred_scores = subset['y_pred_scores']

                sensitive_features = self.get_sensitive_features()
                metrics = self.load_policy()
                grader = Grader(references = y_true, predictions = y_pred, sensitive_features = sensitive_features, policy = self.policy, metrics = metrics)
                result = grader.compute()
                values_subsets_all[key] = result

            return values_subsets_all

    def give_subsets(self):
        subsets = []
        unique_values = np.unique(self.sensitive_features)

        for value in unique_values:
            ind = np.where(self.sensitive_features == value)
            sub_dict = {
                "key": value,
                "y_true": self.y_true[ind],
                "y_pred": self.y_pred[ind],
                "sensitive_features": self.sensitive_features[ind],
            }
            if self.y_pred_scores is not None:
                sub_dict.update({"y_pred_scores": self.y_pred_scores[ind]})
            else:
                sub_dict.update({"y_pred_scores": self.y_pred_scores})

            subsets.append(sub_dict)

        return subsets


def eval(label_model: str, label_dataset: str, policy: dict):
    evaluator = EvalHelper(label_model=label_model,
                           label_dataset=label_dataset, policy=policy)

    evaluator.load()
    complete = evaluator.evaluate()
    subsets = evaluator.evaluate_subsets()
    labelmodel = label_model.split(':')
    version = labelmodel[1]
    model = labelmodel[0]
    result = {'complete' : complete, 'subsets' : subsets}
    formatted_result = {
            "model" : f"{model}",
            "version" : f"{version}",
            "result" : [result]
    }
    plot_balanced_accuracy(formatted_result)
    return formatted_result

def plot_balanced_accuracy(results_json, orientation='horizontal'):
    """
    Plots the balanced accuracy for Complete, Male, and Female categories
    against a specified threshold.
    
    Parameters:
    - results_json: Dictionary containing the results data
    - orientation: 'horizontal' or 'vertical' to determine the bar graph orientation
    """
    # Extracting data
    results = results_json['result'][0]
    complete_data = float(results['complete']['complete']['fairness_scores']['balanced_accuracy'])
    male_data = float(results['subsets'][1]['fairness_scores']['balanced_accuracy'])
    female_data = float(results['subsets'][2]['fairness_scores']['balanced_accuracy'])
    threshold = float(results['complete']['complete']['fairness_thresholds']['balanced_accuracy'])
    
    # Categories
    categories = ['Complete', 'Male', 'Female']
    
    # Values
    values = [complete_data, male_data, female_data]
    
    # Colors
    colors = ['#1f77b4', '#60a3d9', '#c2e0f9']
    
    # Creating the bar graph
    fig, ax = plt.subplots()
    if orientation == 'horizontal':
        bars = ax.barh(categories, values, color=colors)
        ax.axvline(x=threshold, color='gray', linestyle='--', label=f'Threshold ({threshold})')
        ax.set_xlabel('Balanced Accuracy')
        ax.set_ylabel('Categories')
    else:
        bars = ax.bar(categories, values, color=colors)
        ax.axhline(y=threshold, color='gray', linestyle='--', label=f'Threshold ({threshold})')
        ax.set_ylabel('Balanced Accuracy')
        ax.set_xlabel('Categories')
    
    # Adding labels and title
    ax.set_title('Balanced Accuracy Comparison')
    ax.legend()
    
    # Displaying the bar graph
    plt.show()


import pureml
from pureml.components import get_org_id
import requests
from pureml.cli.auth import get_auth_headers
from pureml.schema import BackendSchema, LogSchema, ConfigKeys, ContentTypeHeader
from .schema import policy_list
from urllib.parse import urljoin
import json


def get_policy_details(policy_name='nyc144'):
    policy = policy_list[policy_name]

    task_type = policy['task_type']
    metrics = policy['metrics']
    sensitive_columns = policy['sensitive_columns']

    return task_type, metrics, sensitive_columns


def evaluate_with_policy(policy_name='nyc144', label_model=None, label_dataset=None):

    task_type, metrics, sensitive_columns = get_policy_details(policy_name)

    metric_values = pureml.eval(task_type=task_type,
                                label_model=label_model,
                                label_dataset=label_dataset,
                                metrics=metrics)

    model_name, model_version = label_model.split(":")
    dataset_name, dataset_version = label_model.split(":")
    result_policy = {
        "policy_details": {
            "name": policy_name
        },
        "model_details": {
            "name": model_name,
            "version": model_version
        },
        "dataset_details": {
            "name": dataset_name,
            "version": dataset_version
        },
        "sensitive_columns": sensitive_columns,
        "metrics": metric_values,
        "ethical_considerations": None,
        "Caveats_and_recommendations": None
    }

    response = post_policy_results(result_policy=result_policy,
                                   model_name=model_name,
                                   model_version=model_version,
                                   policy_name=policy_name)

    return result_policy, response


def get_policy_uuid(policy_name):

    backend_schema = BackendSchema().get_instance()

    url = "policies?policyName={}".format(
        policy_name
    )

    url = urljoin(backend_schema.BASE_URL, url)

    headers = get_auth_headers(content_type=ContentTypeHeader.ALL)

    data = {"policyName": policy_name}

    # data = json.dumps(data)

    response = requests.get(url, data=data, headers=headers)

    if response.ok:
        print(f"[bold green]UUID for policy: ",
              policy_name, " have been fetched!")
        uuid = response.json()['data'][0]["uuid"]
        return uuid

    else:
        print(f"[bold red]UUID for policy: ",
              policy_name, " have not been fetched!")

        return None


def post_policy_results(result_policy, model_name, model_version, policy_name):

    policy_uuid = get_policy_uuid(policy_name=policy_name)
    response = None

    if policy_uuid is not None:

        org_id = get_org_id()

        backend_schema = BackendSchema().get_instance()

        url = "reports?orgId={}&modelName={}&version={}".format(
            org_id, model_name, model_version
        )

        url = urljoin(backend_schema.BASE_URL, url)

        headers = get_auth_headers(content_type=ContentTypeHeader.ALL)

        result_policy = json.dumps(result_policy)
        data = {"data": result_policy, "policy_uuid": policy_uuid}

        data = json.dumps(data)

        response = requests.post(url, data=data, headers=headers)

        if response.ok:
            print(f"[bold green]Policy results have been registered!")

        else:
            print(f"[bold red]Policy results  have not been registered!")

    return response





policy_list = {

    'nyc144': {
        'sensitive_columns': ['sex', 'race'],
        'task_type': ['classification', 'fairness'],
        'metrics':
            {
                'classification':
                [
                    {'name': 'accuracy', 'threshold': 0.6}
                ],
                'fairness':
                [
                    {'name': 'demographic_parity_difference', 'threshold': 0.1}
                ]
        }
    }
}
